import abc
import json
from pathlib import Path
from typing import Any, Dict, MutableMapping, Optional, Union

import numpy as np
import zarr
from loguru import logger
from pydantic import (
    Field,
    PrivateAttr,
    computed_field,
    field_serializer,
    field_validator,
    model_validator,
)
from typing_extensions import Self

from polaris._artifact import BaseArtifactModel
from polaris.dataset._adapters import Adapter
from polaris.dataset._column import ColumnAnnotation
from polaris.dataset.zarr import MemoryMappedDirectoryStore
from polaris.dataset.zarr._utils import load_zarr_group_to_memory
from polaris.hub.polarisfs import PolarisFileSystem
from polaris.utils.dict2html import dict2html
from polaris.utils.errors import InvalidDatasetError
from polaris.utils.types import (
    AccessType,
    DatasetIndex,
    HttpUrlString,
    HubOwner,
    SupportedLicenseType,
    ZarrConflictResolution,
)

# Constants
_CACHE_SUBDIR = "datasets"


class BaseDataset(BaseArtifactModel, abc.ABC):
    """Base data-model for a Polaris dataset, implemented as a [Pydantic](https://docs.pydantic.dev/latest/) model.

    At its core, a dataset in Polaris can _conceptually_ be thought of as tabular data structure that stores data-points
    in a row-wise manner, where each column correspond to a variable associated with that datapoint.

    A Dataset can have multiple modalities or targets, can be sparse and can be part of one or multiple
     [`BenchmarkSpecification`][polaris.benchmark.BenchmarkSpecification] objects.

    Attributes:
        default_adapters: The adapters that the Dataset recommends to use by default to change the format of the data
            for specific columns.
        zarr_root_path: The data for any pointer column should be saved in the Zarr archive this path points to.
        readme: Markdown text that can be used to provide a formatted description of the dataset.
            If using the Polaris Hub, it is worth noting that this field is more easily edited through the Hub UI
            as it provides a rich text editor for writing markdown.
        annotations: Each column _can be_ annotated with a [`ColumnAnnotation`][polaris.dataset.ColumnAnnotation] object.
            Importantly, this is used to annotate whether a column is a pointer column.
        source: The data source, e.g. a DOI, Github repo or URI.
        license: The dataset license. Polaris only supports some Creative Commons licenses. See [`SupportedLicenseType`][polaris.utils.types.SupportedLicenseType] for accepted ID values.
        curation_reference: A reference to the curation process, e.g. a DOI, Github repo or URI.
        cache_dir: Where the dataset would be cached if you call the `cache()` method.
    For additional meta-data attributes, see the [`BaseArtifactModel`][polaris._artifact.BaseArtifactModel] class.

    Raises:
        InvalidDatasetError: If the dataset does not conform to the Pydantic data-model specification.
    """

    # Public attributes
    # Data
    default_adapters: Dict[str, Adapter] = Field(default_factory=dict)
    zarr_root_path: Optional[str] = None

    # Additional meta-data
    readme: str = ""
    annotations: Dict[str, ColumnAnnotation] = Field(default_factory=dict)
    source: Optional[HttpUrlString] = None
    license: Optional[SupportedLicenseType] = None
    curation_reference: Optional[HttpUrlString] = None

    # Config
    cache_dir: Optional[Path] = None

    # Private attributes
    _zarr_root: Optional[zarr.Group] = PrivateAttr(None)
    _zarr_data: Optional[MutableMapping[str, np.ndarray]] = PrivateAttr(None)
    _client = PrivateAttr(None)  # Optional[PolarisHubClient]
    _warn_about_remote_zarr: bool = PrivateAttr(True)

    @field_validator("default_adapters", mode="before")
    def _validate_adapters(cls, value):
        """Validate the adapters"""
        return {k: Adapter[v] if isinstance(v, str) else v for k, v in value.items()}

    @field_serializer("default_adapters")
    def _serialize_adapters(self, value: dict[str, Adapter]):
        """Serializes the adapters"""
        return {k: v.name for k, v in value.items()}

    @field_serializer("cache_dir", "zarr_root_path")
    def _serialize_paths(value):
        """Serialize the paths"""
        if value is not None:
            value = str(value)
        return value

    @model_validator(mode="after")
    def _validate_base_dataset_model(self) -> Self:
        # Verify that all annotations are for columns that exist
        if any(k not in self.columns for k in self.annotations):
            raise InvalidDatasetError(
                f"There are annotations for columns that do not exist. Columns: {self.columns}. Annotations: {list(self.annotations.keys())}"
            )

        # Verify that all adapters are for columns that exist
        if any(k not in self.columns for k in self.default_adapters.keys()):
            raise InvalidDatasetError(
                f"There are default adapters for columns that do not exist. Columns: {self.columns}. Adapters: {list(self.annotations.keys())}"
            )

        # Set a default for missing annotations and convert strings to Modality
        for c in self.columns:
            if c not in self.annotations:
                self.annotations[c] = ColumnAnnotation()
            self.annotations[c].dtype = self.dtypes[c]

        return self

    @property
    def client(self):
        """The Polaris Hub client used to interact with the Polaris Hub."""

        # Import it here to prevent circular imports
        from polaris.hub.client import PolarisHubClient

        if self._client is None:
            self._client = PolarisHubClient()
        return self._client

    @property
    def uses_zarr(self) -> bool:
        """Whether any of the data in this dataset is stored in a Zarr Archive."""
        return self.zarr_root_path is not None

    @property
    def zarr_data(self):
        """Get the Zarr data.

        This is different from the Zarr Root, because to optimize the efficiency of
        data loading, a user can choose to load the data into memory as a numpy array

        Note: General purpose dataloader.
            The goal with Polaris is to provide general purpose datasets that serve as good
            options for a _wide variety of use cases_. This also implies you should be able to
            optimize things further for a specific use case if needed.
        """
        if self._zarr_data is not None:
            return self._zarr_data
        return self.zarr_root

    @property
    def zarr_root(self) -> zarr.Group | None:
        """Get the zarr Group object corresponding to the root.

        Opens the zarr archive in read-write mode if it is not already open.

        Note: Different to `zarr_data`
            The `zarr_data` attribute references either to the Zarr archive or to a in-memory copy of the data.
            See also [`Dataset.load_to_memory`][polaris.dataset.Dataset.load_to_memory].
        """

        if self._zarr_root is not None:
            return self._zarr_root

        if self.zarr_root_path is None:
            return None

        # We open the archive in read-only mode if it is saved on the Hub
        saved_on_hub = PolarisFileSystem.is_polarisfs_path(self.zarr_root_path)

        if self._warn_about_remote_zarr:
            saved_remote = saved_on_hub or not Path(self.zarr_root_path).exists()

            if saved_remote:
                logger.warning(
                    f"You're loading data from a remote location. "
                    f"To speed up this process, consider caching the dataset first "
                    f"using {self.__class__.__name__}.cache()"
                )
                self._warn_about_remote_zarr = False

        try:
            if saved_on_hub:
                self._zarr_root = self.client.open_zarr_file(self.owner, self.name, self.zarr_root_path, "r+")
            else:
                # We use memory mapping by default because our experiments show that it's consistently faster
                store = MemoryMappedDirectoryStore(self.zarr_root_path)
                self._zarr_root = zarr.open_consolidated(store, mode="r+")
        except KeyError as error:
            raise InvalidDatasetError(
                "A Zarr archive associated with a Polaris dataset has to be consolidated."
            ) from error
        return self._zarr_root

    @computed_field
    @property
    def n_rows(self) -> int:
        """The number of rows in the dataset."""
        return len(self.rows)

    @computed_field
    @property
    def n_columns(self) -> int:
        """The number of columns in the dataset."""
        return len(self.columns)

    @property
    @abc.abstractmethod
    def rows(self) -> list[str | int]:
        """Return all row indices for the dataset"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def columns(self) -> list[str]:
        """Return all columns for the dataset"""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def dtypes(self) -> dict[str, np.dtype]:
        """Return the dtype for each of the columns for the dataset"""
        raise NotImplementedError

    def load_to_memory(self):
        """
        Load data from zarr files to memeory

        Warning: Make sure the uncompressed dataset fits in-memory.
            This method will load the **uncompressed** dataset into memory. Make
            sure you actually have enough memory to store the dataset.
        """
        data = self.zarr_data

        if not isinstance(data, zarr.Group):
            raise TypeError(
                "The dataset zarr_root is not a valid Zarr archive. "
                "Did you call Dataset.load_to_memory() twice?"
            )

        # NOTE (cwognum): If the dataset fits in memory, the most performant is to use plain NumPy arrays.
        # Even if we disable chunking and compression in Zarr.
        # For more information, see https://github.com/zarr-developers/zarr-python/issues/1395
        self._zarr_data = load_zarr_group_to_memory(data)

    @abc.abstractmethod
    def get_data(
        self, row: str | int, col: str, adapters: dict[str, Adapter] | None = None
    ) -> np.ndarray | Any:
        """Since the dataset might contain pointers to external files, data retrieval is more complicated
        than just indexing the `table` attribute. This method provides an end-point for seamlessly
        accessing the underlying data.

        Args:
            row: The row index in the `Dataset.table` attribute
            col: The column index in the `Dataset.table` attribute
            adapters: The adapters to apply to the data before returning it.
                If None, will use the default adapters specified for the dataset.

        Returns:
            A numpy array with the data at the specified indices. If the column is a pointer column,
                the content of the referenced file is loaded to memory.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def upload_to_hub(self, access: AccessType = "private", owner: Union[HubOwner, str, None] = None):
        """Uploads the dataset to the Polaris Hub."""
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def from_json(cls, path: str):
        """
        Loads a dataset from a JSON file.

        Args:
            path: The path to the JSON file to load the dataset from.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def to_json(
        self,
        destination: str,
        if_exists: ZarrConflictResolution = "replace",
        load_zarr_from_new_location: bool = False,
    ) -> str:
        """
        Save the dataset to a destination directory as a JSON file.

        Args:
            destination: The _directory_ to save the associated data to.
            if_exists: Action for handling existing files in the Zarr archive. Options are 'raise' to throw
                an error, 'replace' to overwrite, or 'skip' to proceed without altering the existing files.
            load_zarr_from_new_location: Whether to update the current instance to load data from the location
                the data is saved to. Only relevant for Zarr-datasets.

        Returns:
            The path to the JSON file.
        """
        raise NotImplementedError

    def cache(self) -> str:
        """Caches the dataset by downloading all additional data for pointer columns to a local directory.

        Returns:
            The path to the cache directory.
        """
        self.to_json(self.cache_dir, load_zarr_from_new_location=True)
        return self.cache_dir

    def size(self) -> tuple[int, int]:
        return self.n_rows, self.n_columns

    def __getitem__(self, item: DatasetIndex) -> Any | np.ndarray | dict[str, np.ndarray]:
        """Allows for indexing the dataset directly"""

        # If a tuple, we assume it's the row and column index pair
        if isinstance(item, tuple):
            row, col = item
            return self.get_data(row, col)

        # Otherwise, we assume you're indexing the row
        return {col: self.get_data(item, col) for col in self.columns}

    @abc.abstractmethod
    def _repr_dict_(self) -> dict:
        """Utility function for pretty-printing to the command line and jupyter notebooks"""
        raise NotImplementedError

    def _repr_html_(self):
        """For pretty-printing in Jupyter Notebooks"""
        return dict2html(self._repr_dict_())

    def __len__(self):
        return self.n_rows

    def __repr__(self):
        return json.dumps(self._repr_dict_(), indent=2)

    def __str__(self):
        return self.__repr__()

    def __del__(self):
        """Close the connection of the client"""
        if self._client is not None:
            self._client.close()



import os
from hashlib import md5

import pyarrow as pa
import pyarrow.parquet as pq

# PyArrow table schema for the V2 Zarr manifest file
ZARR_MANIFEST_SCHEMA = pa.schema([("path", pa.string()), ("checksum", pa.string())])


def generate_zarr_manifest(zarr_root_path: str, output_dir: str):
    """
    Entry point function which triggers the creation of a Zarr manifest for a V2 dataset.

    Parameters:
        zarr_root_path: The path to the root of a Zarr archive
        output_dir: The path to the directory which will hold the generated manifest
    """

    zarr_manifest_path = f"{output_dir}/zarr_manifest.parquet"

    with pq.ParquetWriter(zarr_manifest_path, ZARR_MANIFEST_SCHEMA) as writer:
        recursively_build_manifest(zarr_root_path, writer, zarr_root_path)

    return zarr_manifest_path


def recursively_build_manifest(dir_path: str, writer: pq.ParquetWriter, zarr_root_path: str) -> str:
    """
    Recursive function that traverses a Zarr archive to build a V2 manifest file.

    Parameters:
        dir_path: The path to the current directory being processed in the archive
        writer: Writer object for incrementally adding rows to the manifest Parquet file
        zarr_root_path: The root path which triggered the first recursive call
    """

    # Get iterator of items located in the directory at `dir_path`
    with os.scandir(dir_path) as it:
        #
        # Loop through directory items in iterator
        for entry in it:
            if entry.is_dir():
                #
                # If item is a directory, recurse into that directory
                recursively_build_manifest(entry.path, writer, zarr_root_path)
            elif entry.is_file():
                #
                # If item is a file, calculate its relative path and chunk checksum. Then, append that
                # to the Zarr manifest parquet.
                table = pa.Table.from_pydict(
                    {
                        "path": [os.path.relpath(entry.path, zarr_root_path)],
                        "checksum": [calculate_file_md5(entry.path)],
                    },
                    schema=ZARR_MANIFEST_SCHEMA,
                )
                writer.write_table(table)


def calculate_file_md5(file_path: str):
    """Calculates the md5 hash for a file at a given path"""

    md5_hash = md5()
    with open(file_path, "rb") as file:
        #
        # Read the file in chunks to avoid using too much memory for large files
        for chunk in iter(lambda: file.read(4096), b""):
            md5_hash.update(chunk)

    # Return the hex representation of the digest
    return md5_hash.hexdigest()



import datamol as dm
import numpy as np
import pytest

from polaris.dataset import Subset
from polaris.utils.errors import TestAccessError


def test_consistency_across_access_methods(test_dataset):
    """Using the various endpoints of the Subset API should not lead to the same data."""
    indices = list(range(5))
    task = Subset(test_dataset, indices, "smiles", "expt")

    # Ground truth
    expected_smiles = test_dataset.table.loc[indices, "smiles"]
    expected_targets = test_dataset.table.loc[indices, "expt"]

    # Indexing
    assert ([task[i][0] for i in range(5)] == expected_smiles).all()
    assert ([task[i][1] for i in range(5)] == expected_targets).all()

    # Iterator
    assert (list(smi for smi, y in task) == expected_smiles).all()
    assert (list(y for smi, y in task) == expected_targets).all()

    # Property
    assert (task.inputs == expected_smiles).all()
    assert (task.targets == expected_targets).all()
    assert (task.X == expected_smiles).all()
    assert (task.y == expected_targets).all()


def test_access_to_test_set(test_single_task_benchmark):
    """A user should not have access to the test set targets."""

    train, test = test_single_task_benchmark.get_train_test_split()
    assert test._hide_targets
    assert not train._hide_targets

    with pytest.raises(TestAccessError):
        test.as_array("y")
    with pytest.raises(TestAccessError):
        test.targets

    # Check if iterable style access returns just the SMILES
    for x in test:
        assert isinstance(x, str)
    for i in range(len(test)):
        assert isinstance(test[i], str)

    # For the train set it should work
    assert all(isinstance(y, float) for x, y in train)
    assert all(isinstance(train[i][1], float) for i in range(len(train)))


def test_input_featurization(test_single_task_benchmark):
    # Without a transformation, we expect a SMILES string
    train, test = test_single_task_benchmark.get_train_test_split()
    test_single_task_benchmark._n_splits_since_evaluate = 0  # Manually reset for sake of test

    x, y = train[0]
    assert isinstance(x, str)

    x = test[0]
    assert isinstance(x, str)

    train, test = test_single_task_benchmark.get_train_test_split(featurization_fn=dm.to_fp)

    # For all different flavours of accessing the data
    # Make sure the input is now featurized
    x, y = train[0]
    assert isinstance(x, np.ndarray)

    x = test[0]
    assert isinstance(x, np.ndarray)

    x, y = next(train)
    assert isinstance(x, np.ndarray)

    x = next(test)
    assert isinstance(x, np.ndarray)

    x = train.X[0]
    assert isinstance(x, np.ndarray)

    x = test.X[0]
    assert isinstance(x, np.ndarray)



import enum
from typing import Dict, Literal, Optional, TypeAlias

import numpy as np
from numpy.typing import DTypeLike
from pydantic import BaseModel, ConfigDict, Field, field_serializer, field_validator
from pydantic.alias_generators import to_camel


class Modality(enum.Enum):
    """Used to Annotate columns in a dataset."""

    UNKNOWN = "unknown"
    MOLECULE = "molecule"
    MOLECULE_3D = "molecule_3D"
    PROTEIN = "protein"
    PROTEIN_3D = "protein_3D"
    IMAGE = "image"


KnownContentType: TypeAlias = Literal["chemical/x-smiles", "chemical/x-pdb"]


class ColumnAnnotation(BaseModel):
    """
    The `ColumnAnnotation` class is used to annotate the columns of the [`Dataset`][polaris.dataset.Dataset] object.
    This mostly just stores meta-data and does not affect the logic. The exception is the `is_pointer` attribute.

    Attributes:
        is_pointer: Annotates whether a column is a pointer column. If so, it does not contain data,
            but rather contains references to blobs of data from which the data is loaded.
        modality: The data modality describes the data type and is used to categorize datasets on the hub
            and while it does not affect logic in this library, it does affect the logic of the hub.
        description: Describes how the data was generated.
        user_attributes: Any additional meta-data can be stored in the user attributes.
        content_type: Specify column's IANA content type. If the the content type matches with a known type for
            molecules (e.g. "chemical/x-smiles"), visualization for its content will be activated on the Hub side
    """

    is_pointer: bool = False
    modality: Modality = Modality.UNKNOWN
    description: Optional[str] = None
    user_attributes: Dict[str, str] = Field(default_factory=dict)
    dtype: np.dtype | None = None
    content_type: KnownContentType | str | None = None

    model_config = ConfigDict(arbitrary_types_allowed=True, alias_generator=to_camel, populate_by_name=True)

    @field_validator("modality", mode="before")
    def _validate_modality(cls, v):
        """Tries to convert a string to the Enum"""
        if isinstance(v, str):
            v = Modality[v.upper()]
        return v

    @field_validator("dtype", mode="before")
    def _validate_dtype(cls, v):
        """Tries to convert a string to the Enum"""
        if isinstance(v, str):
            v = np.dtype(v)
        return v

    @field_serializer("modality")
    def _serialize_modality(self, v: Modality):
        """Return the modality as a string, keeping it serializable"""
        return v.name

    @field_serializer("dtype")
    def _serialize_dtype(self, v: Optional[DTypeLike]):
        """Return the dtype as a string, keeping it serializable"""
        if v is not None:
            v = v.name
        return v



from polaris.evaluate._metric import Metric, MetricInfo
from polaris.evaluate._results import (
    BenchmarkResults,
    ResultsType,
    CompetitionResults,
    CompetitionPredictions,
    ResultsMetadata,
    EvaluationResult,
)
from polaris.evaluate.utils import evaluate_benchmark

__all__ = [
    "Metric",
    "MetricInfo",
    "ResultsMetadata",
    "EvaluationResult",
    "BenchmarkResults",
    "CompetitionResults",
    "ResultsType",
    "evaluate_benchmark",
    "CompetitionPredictions",
]



from collections import defaultdict
from typing import TYPE_CHECKING
import os

import pandas as pd
import zarr

from polaris.dataset import ColumnAnnotation
from polaris.dataset.converters._base import Converter, FactoryProduct

if TYPE_CHECKING:
    from polaris.dataset import DatasetFactory


class ZarrConverter(Converter):
    """Parse a [.zarr](https://zarr.readthedocs.io/en/stable/index.html) archive into a Polaris `Dataset`.

    Tip: Tutorial
        To learn more about the zarr format, see the
        [tutorial](../tutorials/dataset_zarr.ipynb).

    Warning: Loading from `.zarr`
        Loading and saving datasets from and to `.zarr` is still experimental and currently not
        fully supported by the Hub.

    A `.zarr` file can contain groups and arrays, where each group can again contain groups and arrays.
    Within Polaris, the Zarr archive is expected to have a flat hierarchy where each array corresponds
    to a single column and each array contains the values for all datapoints in that column.
    """

    def convert(self, path: str, factory: "DatasetFactory", append: bool = False) -> FactoryProduct:
        src = zarr.open(path, "r")

        v = next(src.group_keys(), None)
        if v is not None:
            raise ValueError("The root of the zarr hierarchy should only contain arrays.")

        # Copy to the source zarr, so everything is in one place
        pointer_start_dict = {col: 0 for col, _ in src.arrays()}
        if append:
            if not os.path.exists(factory.zarr_root.store.path):
                raise RuntimeError(f"Zarr store {factory.zarr_root.store.path} doesn't exist. \
                    Please make sure the zarr store {factory.zarr_root.store.path} is created. Or set `append` to `False`.")
            else:
                for col, arr in src.arrays():
                    pointer_start_dict[col] += factory.zarr_root[col].shape[0]
                    factory.zarr_root[col].append(arr)
        else:
            zarr.copy_store(source=src.store, dest=factory.zarr_root.store, if_exists="skip")

        # Construct the table
        # Parse any group into a column
        data = defaultdict(dict)
        for col, arr in src.arrays():
            for i in range(len(arr)):
                data[col][i] = self.get_pointer(arr.name.removeprefix("/"), i)

        # Construct the dataset
        table = pd.DataFrame(data)
        return table, {k: ColumnAnnotation(is_pointer=True) for k in table.columns}, {}



from polaris.benchmark._base import BenchmarkSpecification
from polaris.benchmark._definitions import (
    MultiTaskBenchmarkSpecification,
    SingleTaskBenchmarkSpecification,
)

__all__ = ["BenchmarkSpecification", "SingleTaskBenchmarkSpecification", "MultiTaskBenchmarkSpecification"]



import json

from polaris.hub.oauth import CachedTokenAuth


def test_cached_token_auth_empty_on_no_cache(tmp_path):
    filename = "test_token.json"
    auth = CachedTokenAuth(cache_dir=tmp_path, filename=filename)
    assert not auth.token


def test_cached_token_auth_reads_from_cache(tmp_path):
    filename = "test_token.json"
    cache_file = tmp_path / filename
    cache_file.write_text(
        json.dumps(
            {
                "access_token": "test_token",
                "issued_token_type": "urn:ietf:params:oauth:token-type:jwt",
                "token_type": "Bearer",
                "expires_in": 576618,
                "expires_at": 1720122005,
            }
        )
    )

    auth = CachedTokenAuth(cache_dir=tmp_path, filename=filename)

    assert auth.token is not None
    assert auth.token["access_token"] == "test_token"
    assert auth.token["expires_at"] == 1720122005
    assert auth.token["expires_in"] == 576618
    assert auth.token["token_type"] == "Bearer"
    assert auth.token["issued_token_type"] == "urn:ietf:params:oauth:token-type:jwt"


def test_cached_token_auth_writes_to_cache(tmp_path):
    filename = "test_token.json"
    cache_file = tmp_path / filename

    auth = CachedTokenAuth(cache_dir=tmp_path, filename=filename)
    auth.set_token(
        {
            "access_token": "test_token",
            "issued_token_type": "urn:ietf:params:oauth:token-type:jwt",
            "token_type": "Bearer",
            "expires_in": 576618,
            "expires_at": 1720122005,
        }
    )

    assert cache_file.exists()
    assert json.loads(cache_file.read_text()) == auth.token



from polaris.dataset.converters._base import Converter
from polaris.dataset.converters._sdf import SDFConverter
from polaris.dataset.converters._zarr import ZarrConverter
from polaris.dataset.converters._pdb import PDBConverter


__all__ = ["Converter", "SDFConverter", "ZarrConverter", "PDBConverter"]



import numpy as np
import pandas as pd

from polaris.evaluate.utils import evaluate_benchmark, normalize_predictions_type
from polaris.competition import CompetitionSpecification


def test_competition_from_json(test_competition, tmpdir):
    """Test whether we can successfully save and load a competition from JSON."""
    path = test_competition.to_json(str(tmpdir))
    new_competition = CompetitionSpecification.from_json(path)
    assert new_competition == test_competition


def test_multi_col_competition_evaluation(test_competition):
    """Test that multi-column competitions will be evaluated properly when when
    target labels are read as a pandas dataframe from a file."""
    data = np.random.randint(2, size=(6, 3))
    labels = pd.DataFrame(data, columns=["Column1", "Column2", "Column3"])
    labels_as_from_hub = {col: np.array(labels[col]) for col in labels.columns}
    predictions = {target_col: np.random.randint(2, size=labels.shape[0]) for target_col in labels.columns}

    result = evaluate_benchmark(
        ["Column1", "Column2", "Column3"], test_competition.metrics, labels_as_from_hub, y_pred=predictions
    )

    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == {
        "Test set",
        "Target label",
        "Metric",
        "Score",
    }


def test_single_col_competition_evaluation(test_competition):
    """Test that multi-column competitions will be evaluated properly when when
    target labels are read as a pandas dataframe from a file."""
    data = np.array(
        [
            1.15588236,
            1.56414507,
            1.04828639,
            0.98362629,
            1.22613572,
            2.56594576,
            0.67568671,
            0.86099644,
            0.67568671,
            2.28213589,
            1.06617679,
            1.05709529,
            0.67568671,
            0.67568671,
            0.67568671,
        ]
    )
    labels = {"LOG HLM_CLint (mL/min/kg)": data}
    predictions = data + np.random.uniform(0, 3, size=len(data))

    result = evaluate_benchmark(
        ["LOG HLM_CLint (mL/min/kg)"], test_competition.metrics, labels, y_pred=predictions
    )

    assert isinstance(result, pd.DataFrame)
    assert set(result.columns) == {
        "Test set",
        "Target label",
        "Metric",
        "Score",
    }


def test_normalize_predictions_type():
    "Single column, single test set"
    assert {"test": {"col1": [1, 2, 3]}} == normalize_predictions_type([1, 2, 3], ["col1"])
    assert {"test": {"col1": [1, 2, 3]}} == normalize_predictions_type({"col1": [1, 2, 3]}, ["col1"])
    assert {"test": {"col1": [1, 2, 3]}} == normalize_predictions_type(
        {"test": {"col1": [1, 2, 3]}}, ["col1"]
    )

    "Multi-column, single test set"
    assert {"test": {"col1": [1, 2, 3], "col2": [4, 5, 6]}} == normalize_predictions_type(
        {"col1": [1, 2, 3], "col2": [4, 5, 6]}, ["col1", "col2"]
    )

    assert {"test": {"col1": [1, 2, 3], "col2": [4, 5, 6]}} == normalize_predictions_type(
        {"test": {"col1": [1, 2, 3], "col2": [4, 5, 6]}}, ["col1", "col2"]
    )

    "Single column, multi-test set"
    assert {"test1": {"col1": [1, 2, 3]}, "test2": {"col1": [4, 5, 6]}} == normalize_predictions_type(
        {"test1": {"col1": [1, 2, 3]}, "test2": {"col1": [4, 5, 6]}}, ["col1"]
    )

    "Multi-column, multi-test set"
    assert {
        "test1": {"col1": [1, 2, 3], "col2": [4, 5, 6]},
        "test2": {"col1": [7, 8, 9], "col2": [10, 11, 12]},
    } == normalize_predictions_type(
        {"test1": {"col1": [1, 2, 3], "col2": [4, 5, 6]}, "test2": {"col1": [7, 8, 9], "col2": [10, 11, 12]}},
        ["col1", "col2"],
    )



from polaris.dataset._column import ColumnAnnotation, KnownContentType, Modality
from polaris.dataset._competition_dataset import CompetitionDataset
from polaris.dataset._dataset import DatasetV1
from polaris.dataset._dataset import DatasetV1 as Dataset
from polaris.dataset._factory import DatasetFactory, create_dataset_from_file, create_dataset_from_files
from polaris.dataset._subset import Subset

__all__ = [
    "ColumnAnnotation",
    "Dataset",
    "CompetitionDataset",
    "Subset",
    "Modality",
    "KnownContentType",
    "DatasetFactory",
    "create_dataset_from_file",
    "create_dataset_from_files",
    "DatasetV1",
]



from datetime import datetime
from typing import Optional

from polaris.benchmark import BenchmarkSpecification
from polaris.evaluate._results import CompetitionPredictions
from polaris.hub.settings import PolarisHubSettings
from polaris.utils.types import HubOwner


class CompetitionSpecification(BenchmarkSpecification):
    """Much of the underlying data model and logic is shared across Benchmarks and Competitions, and
    anything within this class serves as a point of differentiation between the two.

    Attributes:
        owner: A slug-compatible name for the owner of the competition. This is redefined such
            that it is required.
        start_time: The time at which the competition becomes active and interactable.
        end_time: The time at which the competition ends and is no longer interactable.
    """

    # Additional properties specific to Competitions
    owner: HubOwner
    start_time: datetime | None = None
    end_time: datetime | None = None

    def evaluate(
        self,
        predictions: CompetitionPredictions,
        settings: Optional[PolarisHubSettings] = None,
        cache_auth_token: bool = True,
        **kwargs: dict,
    ):
        """Light convenience wrapper around
        [`PolarisHubClient.evaluate_competition`][polaris.hub.client.PolarisHubClient.evaluate_competition].
        """
        from polaris.hub.client import PolarisHubClient

        with PolarisHubClient(
            settings=settings,
            cache_auth_token=cache_auth_token,
            **kwargs,
        ) as client:
            return client.evaluate_competition(self, predictions)



from enum import Enum, auto, unique

import datamol as dm


@unique
class Adapter(Enum):
    """
    Adapters are predefined callables that change the format of the data.
    Adapters are serializable and can thus be saved alongside datasets.

    Attributes:
        SMILES_TO_MOL: Convert a SMILES string to a RDKit molecule.
        BYTES_TO_MOL: Convert a RDKit binary string to a RDKit molecule.
        ARRAY_TO_PDB: Convert a Zarr arrays to PDB arrays.
    """

    SMILES_TO_MOL = auto()
    BYTES_TO_MOL = auto()
    ARRAY_TO_PDB = auto()

    def __call__(self, data):
        # Import here to prevent a cyclic import
        # Given the close coupling between `zarr_to_pdb` and the PDB converter,
        # we wanted to keep those functions in one file which was leading to a cyclic import.
        from polaris.dataset.converters._pdb import zarr_to_pdb

        conversion_map = {"SMILES_TO_MOL": dm.to_mol, "BYTES_TO_MOL": dm.Mol, "ARRAY_TO_PDB": zarr_to_pdb}

        if isinstance(data, tuple):
            return tuple(conversion_map[self.name](d) for d in data)
        return conversion_map[self.name](data)



import datamol as dm
import pandas as pd
from zarr.errors import ContainsArrayError
import pytest
from fastpdb import struc

from polaris.dataset import DatasetFactory, create_dataset_from_file, create_dataset_from_files
from polaris.dataset.converters import PDBConverter, SDFConverter, ZarrConverter


def _check_pdb_dataset(dataset, ground_truth):
    assert len(dataset) == len(ground_truth)
    for i in range(dataset.table.shape[0]):
        pdb_array = dataset.get_data(row=i, col="pdb")
        assert isinstance(pdb_array, struc.AtomArray)
        assert pdb_array[0] == ground_truth[i][0]
        assert pdb_array.equal_annotations(ground_truth[i])


def _check_dataset(dataset, ground_truth, mol_props_as_col):
    assert len(dataset) == len(ground_truth)

    for row in range(len(dataset)):
        mol = dataset.get_data(row=row, col="molecule")

        assert isinstance(mol, dm.Mol)

        if mol_props_as_col:
            assert not mol.HasProp("my_property")
            v = dataset.get_data(row=row, col="my_property")
            assert v == ground_truth[row].GetProp("my_property")

        else:
            assert mol.HasProp("my_property")
            assert mol.GetProp("my_property") == ground_truth[row].GetProp("my_property")
            assert "my_property" not in dataset.columns


def test_sdf_zarr_conversion(sdf_file, caffeine, tmpdir):
    """Test conversion between SDF and Zarr with utility function"""
    dataset = create_dataset_from_file(sdf_file, tmpdir.join("archive.zarr"))
    _check_dataset(dataset, [caffeine], True)


@pytest.mark.parametrize("mol_props_as_col", [True, False])
def test_factory_sdf_with_prop_as_col(sdf_file, caffeine, tmpdir, mol_props_as_col):
    """Test conversion between SDF and Zarr with factory pattern"""

    factory = DatasetFactory(tmpdir.join("archive.zarr"))

    converter = SDFConverter(mol_prop_as_cols=mol_props_as_col)
    factory.register_converter("sdf", converter)

    factory.add_from_file(sdf_file)
    dataset = factory.build()

    _check_dataset(dataset, [caffeine], mol_props_as_col)


def test_zarr_to_zarr_conversion(zarr_archive, tmpdir):
    """Test conversion between Zarr and Zarr with utility function"""
    dataset = create_dataset_from_file(zarr_archive, tmpdir.join("archive.zarr"))
    assert len(dataset) == 100
    assert len(dataset.columns) == 2
    assert all(c in dataset.columns for c in ["A", "B"])
    assert all(dataset.annotations[c].is_pointer for c in ["A", "B"])
    assert dataset.get_data(row=0, col="A").shape == (2048,)


def test_zarr_with_factory_pattern(zarr_archive, tmpdir):
    """Test conversion between Zarr and Zarr with factory pattern"""

    factory = DatasetFactory(tmpdir.join("archive.zarr"))
    converter = ZarrConverter()
    factory.register_converter("zarr", converter)
    factory.add_from_file(zarr_archive)

    factory.add_column(pd.Series([1, 2, 3, 4] * 25, name="C"))

    df = pd.DataFrame({"C": [1, 2, 3, 4], "D": ["W", "X", "Y", "Z"]})
    factory.add_columns(df, merge_on="C")

    dataset = factory.build()
    assert len(dataset) == 100
    assert len(dataset.columns) == 4
    assert all(c in dataset.columns for c in ["A", "B", "C", "D"])
    assert dataset.table["C"].apply({1: "W", 2: "X", 3: "Y", 4: "Z"}.get).equals(dataset.table["D"])


def test_factory_pdb(pdbs_structs, pdb_paths, tmpdir):
    """Test conversion between PDB file and Zarr with factory pattern"""
    factory = DatasetFactory(tmpdir.join("pdb.zarr"))

    converter = PDBConverter()
    factory.register_converter("pdb", converter)

    factory.add_from_file(pdb_paths[0])
    dataset = factory.build()

    _check_pdb_dataset(dataset, pdbs_structs[:1])


def test_factory_pdbs(pdbs_structs, pdb_paths, tmpdir):
    """Test conversion between PDB files and Zarr with factory pattern"""

    factory = DatasetFactory(tmpdir.join("pdbs.zarr"))

    converter = PDBConverter()
    factory.register_converter("pdb", converter)

    factory.add_from_files(pdb_paths, axis=0)
    dataset = factory.build()

    assert dataset.table.shape[0] == len(pdb_paths)
    _check_pdb_dataset(dataset, pdbs_structs)


def test_pdbs_zarr_conversion(pdbs_structs, pdb_paths, tmpdir):
    """Test conversion between PDBs and Zarr with utility function"""

    dataset = create_dataset_from_files(pdb_paths, tmpdir.join("pdbs_2.zarr"), axis=0)

    assert dataset.table.shape[0] == len(pdb_paths)
    _check_pdb_dataset(dataset, pdbs_structs)


def test_factory_sdfs(sdf_files, caffeine, ibuprofen, tmpdir):
    """Test conversion between SDF and Zarr with factory pattern"""

    factory = DatasetFactory(tmpdir.join("sdfs.zarr"))

    converter = SDFConverter(mol_prop_as_cols=True)
    factory.register_converter("sdf", converter)

    factory.add_from_files(sdf_files, axis=0)
    dataset = factory.build()

    _check_dataset(dataset, [caffeine, ibuprofen], True)


def test_factory_sdf_pdb(sdf_file, pdb_paths, caffeine, pdbs_structs, tmpdir):
    """Test conversion between SDF and PDB from files to Zarr with factory pattern"""

    factory = DatasetFactory(tmpdir.join("sdf_pdb.zarr"))

    sdf_converter = SDFConverter(mol_prop_as_cols=False)
    factory.register_converter("sdf", sdf_converter)

    pdb_converter = PDBConverter()
    factory.register_converter("pdb", pdb_converter)

    factory.add_from_files([sdf_file, pdb_paths[0]], axis=1)
    dataset = factory.build()

    _check_dataset(dataset, [caffeine], False)
    _check_pdb_dataset(dataset, pdbs_structs[:1])


def test_factory_from_files_same_column(sdf_files, pdb_paths, tmpdir):
    factory = DatasetFactory(tmpdir.join("files.zarr"))

    sdf_converter = SDFConverter(mol_prop_as_cols=False)
    factory.register_converter("sdf", sdf_converter)

    pdb_converter = PDBConverter()
    factory.register_converter("pdb", pdb_converter)

    # do not allow same type of files to be appended in columns by `add_from_files`
    # in this case, user should define converter for individual columns

    # attempt to append columns by pdbs
    with pytest.raises(ValueError):
        factory.add_from_files(pdb_paths, axis=1)

    # attempt to append columns by sdfs
    with pytest.raises(ContainsArrayError):
        factory.add_from_files(sdf_files, axis=1)



import numpy as np
import pandas as pd
from typing import Optional

from polaris.evaluate import BenchmarkResults, ResultsType
from polaris.utils.types import PredictionsType
from polaris.evaluate import Metric
from numpy.typing import NDArray


def is_multi_task_single_test_set(vals: PredictionsType, target_cols: list[str]):
    """Check if the given values are for a multiple-task benchmark with a single
    test set. This is inferred by comparing the target names with the keys of the
    given data. If all keys in the given data match the target column names, we
    assume they are target names (as opposed to test set names for a single-task,
    multiple test set benchmark)."""
    return all(k in target_cols for k in vals)


def normalize_predictions_type(vals: PredictionsType, target_cols: list[str]):
    if isinstance(vals, dict):
        if is_multi_task_single_test_set(vals, target_cols):
            return {"test": vals}
        else:
            return vals
    elif vals is None:
        return None
    else:
        return {"test": {target_cols[0]: vals}}


def safe_mask(
    input_values: dict | dict[str, dict], test_label: str, target_label: str, mask: NDArray[np.bool_]
):
    if (
        input_values is None
        or input_values.get(test_label) is None
        or input_values[test_label].get(target_label) is None
    ):
        return None
    else:
        return input_values[test_label][target_label][mask]


def evaluate_benchmark(
    target_cols: list[str],
    metrics: list[Metric],
    y_true: PredictionsType,
    y_pred: Optional[PredictionsType] = None,
    y_prob: Optional[PredictionsType] = None,
):
    y_true = normalize_predictions_type(y_true, target_cols)
    y_pred = normalize_predictions_type(y_pred, target_cols)
    y_prob = normalize_predictions_type(y_prob, target_cols)

    if y_pred and set(y_true.keys()) != set(y_pred.keys()):
        raise KeyError(f"Missing keys for at least one of the test sets. Expecting: {sorted(y_true.keys())}")

    # Results are saved in a tabular format. For more info, see the BenchmarkResults docs.
    scores: ResultsType = pd.DataFrame(columns=BenchmarkResults.RESULTS_COLUMNS)

    # For every test set...
    for test_label, y_true_subset in y_true.items():
        # For every metric...
        for metric in metrics:
            if metric.is_multitask:
                # Multi-task but with a metric across targets
                score = metric(
                    y_true=y_true_subset, y_pred=y_pred.get(test_label), y_prob=y_prob.get(test_label)
                )

                scores.loc[len(scores)] = (test_label, "aggregated", metric, score)
                continue

            if not isinstance(y_true_subset, dict):
                # Single task
                score = metric(
                    y_true=y_true_subset, y_pred=y_pred.get(test_label), y_prob=y_prob.get(test_label)
                )
                scores.loc[len(scores)] = (test_label, target_cols[0], metric, score)
                continue

            # Otherwise, for every target...
            for target_label, y_true_target in y_true_subset.items():
                # Single-task metrics for a multi-task benchmark
                # In such a setting, there can be NaN values, which we thus have to filter out.
                mask = ~np.isnan(y_true_target)
                score = metric(
                    y_true=y_true_target[mask],
                    y_pred=safe_mask(y_pred, test_label, target_label, mask),
                    y_prob=safe_mask(y_prob, test_label, target_label, mask),
                )

                scores.loc[len(scores)] = (test_label, target_label, metric, score)

    return scores



import zarr
import zarr.storage


def load_zarr_group_to_memory(group: zarr.Group) -> dict:
    """Loads an entire Zarr group into memory."""
    data = {}
    for key, item in group.items():
        if isinstance(item, zarr.Array):
            data[key] = item[:]
        elif isinstance(item, zarr.Group):
            data[key] = load_zarr_group_to_memory(item)
    return data



import json
from datetime import datetime
from typing import ClassVar, Optional, Union

import numpy as np
import pandas as pd
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    PrivateAttr,
    computed_field,
    field_serializer,
    field_validator,
)
from pydantic.alias_generators import to_camel

from polaris._artifact import BaseArtifactModel
from polaris.evaluate import Metric
from polaris.hub.settings import PolarisHubSettings
from polaris.utils.dict2html import dict2html
from polaris.utils.errors import InvalidResultError
from polaris.utils.misc import sluggify
from polaris.utils.types import (
    AccessType,
    CompetitionPredictionsType,
    HttpUrlString,
    HubOwner,
    HubUser,
    PredictionsType,
    SlugCompatibleStringType,
)

# Define some helpful type aliases
TestLabelType = str
TargetLabelType = str


class ResultRecords(BaseModel):
    """
    A grouped, tabular data-structure to save the actual data of the benchmark results.
    The grouping by test set and target label is done to make it easier to build leaderboards.
    """

    test_set: TestLabelType
    target_label: TargetLabelType
    scores: dict[Union[Metric, str], float]

    # Model config
    model_config = ConfigDict(alias_generator=to_camel, populate_by_name=True)

    @field_validator("scores")
    def validate_scores(cls, v):
        validated = {}
        for metric, score in v.items():
            if not isinstance(metric, Metric):
                try:
                    metric = Metric[metric]
                except KeyError as error:
                    raise ValueError("Invalid metric name") from error
            validated[metric] = score
        return validated

    @field_serializer("scores")
    def serialize_scores(self, value: dict):
        """Change from the Metric enum to a string representation"""
        return {metric.name: score for metric, score in value.items()}


ResultsType = Union[pd.DataFrame, list[ResultRecords | dict]]


class ResultsMetadata(BaseArtifactModel):
    """Base class for evaluation results

    Attributes:
        github_url: The URL to the GitHub repository of the code used to generate these results.
        paper_url: The URL to the paper describing the methodology used to generate these results.
        contributors: The users that are credited for these results.
        _created_at: The time-stamp at which the results were created. Automatically set.
    For additional meta-data attributes, see the [`BaseArtifactModel`][polaris._artifact.BaseArtifactModel] class.
    """

    # Additional meta-data
    github_url: Optional[HttpUrlString] = None
    paper_url: Optional[HttpUrlString] = None
    contributors: Optional[list[HubUser]] = None

    # Private attributes
    _created_at: datetime = PrivateAttr(default_factory=datetime.now)

    def _repr_dict_(self) -> dict:
        """Utility function for pretty-printing to the command line and jupyter notebooks"""
        repr_dict = self.model_dump(exclude=["results"])

        df = self.results.copy(deep=True)
        df["Metric"] = df["Metric"].apply(lambda x: x.name if isinstance(x, Metric) else x)
        repr_dict["results"] = json.loads(df.to_json(orient="records"))

        return repr_dict

    def _repr_html_(self):
        """For pretty-printing in Jupyter Notebooks"""
        return dict2html(self._repr_dict_())

    def __repr__(self):
        return json.dumps(self._repr_dict_(), indent=2)


class EvaluationResult(ResultsMetadata):
    """Class for saving evaluation results

    The actual results are saved in the `results` field using the following tabular format:

    | Test set | Target label | Metric | Score |
    | -------- | ------------ | ------ | ----- |
    | test_iid | EGFR_WT      | AUC    | 0.9   |
    | test_ood | EGFR_WT      | AUC    | 0.75  |
    |    ...   |      ...     |  ...   |  ...  |
    | test_ood | EGFR_L858R   | AUC    | 0.79  |

    question: Categorizing methods
        An open question is how to best categorize a methodology (e.g. a model).
        This is needed since we would like to be able to aggregate results across benchmarks/competitions too,
        to say something about which (type of) methods performs best _in general_.

    Attributes:
        results: Evaluation results are stored directly in a dataframe or in a serialized, JSON compatible dict
            that can be decoded into the associated tabular format.
    For additional meta-data attributes, see the [`ResultsMetadata`][polaris.evaluate._results.ResultsMetadata] class.
    """

    # Define the columns of the results table
    RESULTS_COLUMNS: ClassVar[list[str]] = ["Test set", "Target label", "Metric", "Score"]

    # Results attribute
    results: ResultsType

    @field_validator("results")
    def _validate_results(cls, v):
        """Ensure the results are a valid dataframe and have the expected columns"""

        # If not a dataframe, assume it is a JSON-serialized export of a dataframe.
        if not isinstance(v, pd.DataFrame):
            try:
                df = pd.DataFrame(columns=cls.RESULTS_COLUMNS)
                for record in v:
                    if isinstance(record, dict):
                        record = ResultRecords(**record)

                    for metric, score in record.scores.items():
                        df.loc[len(df)] = {
                            "Test set": record.test_set,
                            "Target label": record.target_label,
                            "Metric": metric,
                            "Score": score,
                        }
                v = df

            except (ValueError, UnicodeDecodeError) as error:
                raise InvalidResultError(
                    f"The provided dictionary cannot be parsed into a {cls.__name__} instance."
                ) from error

        # Check if the dataframe contains _only_ the expected columns
        if set(v.columns) != set(cls.RESULTS_COLUMNS):
            raise InvalidResultError(
                f"The results dataframe should have the following columns: {cls.RESULTS_COLUMNS}"
            )

        # Check if the results are not empty
        if v.empty:
            raise InvalidResultError("The results dataframe is empty")

        # NOTE (cwognum): Since we have a reference to the benchmark, I considered validating the values in the
        #  columns as well (e.g. are all metrics, targets and test sets actually part of the benchmark).
        #  However, to keep this class light-weight, I did not want to add a strict dependency on the full benchmark class.
        #  Especially because validation will happen on the Hub as well before it is shown there.
        return v

    @field_serializer("results")
    def _serialize_results(self, value: ResultsType):
        """Change from the Metric enum to a string representation"""
        self.results["Metric"] = self.results["Metric"].apply(
            lambda x: x.name if isinstance(x, Metric) else x
        )

        serialized = []
        grouped = self.results.groupby(["Test set", "Target label"])
        for (test_set, target_label), group in grouped:
            metrics = {row["Metric"]: row["Score"] for _, row in group.iterrows()}
            record = ResultRecords(test_set=test_set, target_label=target_label, scores=metrics)
            serialized.append(record.model_dump(by_alias=True))

        return serialized


class BenchmarkResults(EvaluationResult):
    """Class specific to results for standard benchmarks.

    This object is returned by [`BenchmarkSpecification.evaluate`][polaris.benchmark.BenchmarkSpecification.evaluate].
    In addition to the metrics on the test set, it contains additional meta-data and logic to integrate
    the results with the Polaris Hub.

    benchmark_name: The name of the benchmark for which these results were generated.
        Together with the benchmark owner, this uniquely identifies the benchmark on the Hub.
    benchmark_owner: The owner of the benchmark for which these results were generated.
        Together with the benchmark name, this uniquely identifies the benchmark on the Hub.
    """

    benchmark_name: SlugCompatibleStringType = Field(..., frozen=True)
    benchmark_owner: Optional[HubOwner] = Field(None, frozen=True)

    @computed_field
    @property
    def benchmark_artifact_id(self) -> str:
        return f"{self.benchmark_owner}/{sluggify(self.benchmark_name)}"

    def upload_to_hub(
        self,
        settings: Optional[PolarisHubSettings] = None,
        cache_auth_token: bool = True,
        access: Optional[AccessType] = "private",
        owner: Union[HubOwner, str, None] = None,
        **kwargs: dict,
    ):
        """
        Very light, convenient wrapper around the
        [`PolarisHubClient.upload_results`][polaris.hub.client.PolarisHubClient.upload_results] method.
        """
        from polaris.hub.client import PolarisHubClient

        with PolarisHubClient(settings=settings, cache_auth_token=cache_auth_token, **kwargs) as client:
            return client.upload_results(self, access=access, owner=owner)


class CompetitionResults(EvaluationResult):
    """Class specific to results for competition benchmarks.

    This object is returned by [`CompetitionSpecification.evaluate`][polaris.competition.CompetitionSpecification.evaluate].
    In addition to the metrics on the test set, it contains additional meta-data and logic to integrate
    the results with the Polaris Hub.

    Attributes:
        competition_name: The name of the competition for which these results were generated.
            Together with the competition owner, this uniquely identifies the competition on the Hub.
        competition_owner: The owner of the competition for which these results were generated.
            Together with the competition name, this uniquely identifies the competition on the Hub.
    """

    competition_name: SlugCompatibleStringType = Field(..., frozen=True)
    competition_owner: Optional[HubOwner] = Field(None, frozen=True)

    @computed_field
    @property
    def competition_artifact_id(self) -> str:
        return f"{self.competition_owner}/{sluggify(self.competition_name)}"


class CompetitionPredictions(ResultsMetadata):
    """Class specific to predictions for competition benchmarks.

    This object is to be used as input to [`CompetitionSpecification.evaluate`][polaris.competition.CompetitionSpecification.evaluate].
    It is used to ensure that the structure of the predictions are compatible with evaluation methods on the Polaris Hub.

    Attributes:
        predictions: The predictions created for a given competition's test set(s).
    """

    predictions: Union[PredictionsType, CompetitionPredictionsType]
    access: Optional[AccessType] = "private"

    @field_validator("predictions")
    @classmethod
    def _convert_predictions(cls, value: Union[PredictionsType, CompetitionPredictionsType]):
        """Convert prediction arrays from a list type to a numpy array. This is required for certain
        operations during prediction evaluation"""

        if isinstance(value, list):
            return np.array(value)
        elif isinstance(value, np.ndarray):
            return value
        elif isinstance(value, dict):
            for key, val in value.items():
                value[key] = cls._convert_predictions(val)
            return value

    @field_serializer("predictions")
    def _serialize_predictions(self, value: PredictionsType):
        """Used to serialize a Predictions object such that it can be sent over the wire during
        external evaluation for competitions"""

        if isinstance(value, np.ndarray):
            return value.tolist()
        elif isinstance(value, list):
            return value
        elif isinstance(value, dict):
            for key, val in value.items():
                value[key] = self._serialize_predictions(val)
            return value



import json
from hashlib import md5
from itertools import chain
from typing import Any, Callable, Optional, Union

import fsspec
import numpy as np
from datamol.utils import fs
from loguru import logger
from pydantic import (
    Field,
    ValidationInfo,
    computed_field,
    field_serializer,
    field_validator,
    model_validator,
)
from sklearn.utils.multiclass import type_of_target
from typing_extensions import Self

from polaris._artifact import BaseArtifactModel
from polaris.dataset import CompetitionDataset, DatasetV1, Subset
from polaris.evaluate import BenchmarkResults, Metric
from polaris.evaluate.utils import evaluate_benchmark
from polaris.hub.settings import PolarisHubSettings
from polaris.mixins import ChecksumMixin
from polaris.utils.dict2html import dict2html
from polaris.utils.errors import InvalidBenchmarkError
from polaris.utils.misc import listit
from polaris.utils.types import (
    AccessType,
    HubOwner,
    PredictionsType,
    SplitType,
    TargetType,
    TaskType,
)

ColumnsType = Union[str, list[str]]


class BenchmarkSpecification(BaseArtifactModel, ChecksumMixin):
    """This class wraps a [`Dataset`][polaris.dataset.Dataset] with additional data
     to specify the evaluation logic.

    Specifically, it specifies:

    1. Which dataset to use (see [`Dataset`][polaris.dataset.Dataset]);
    2. Which columns are used as input and which columns are used as target;
    3. Which metrics should be used to evaluate performance on this task;
    4. A predefined, static train-test split to use during evaluation.

    info: Subclasses
        Polaris includes various subclasses of the `BenchmarkSpecification` that provide a more precise data-model or
         additional logic, e.g. [`SingleTaskBenchmarkSpecification`][polaris.benchmark.SingleTaskBenchmarkSpecification].

    Examples:
        Basic API usage:
        ```python
        import polaris as po

        # Load the benchmark from the Hub
        benchmark = po.load_benchmark("polaris/hello-world-benchmark")

        # Get the train and test data-loaders
        train, test = benchmark.get_train_test_split()

        # Use the training data to train your model
        # Get the input as an array with 'train.inputs' and 'train.targets'
        # Or simply iterate over the train object.
        for x, y in train:
            ...

        # Work your magic to accurately predict the test set
        predictions = [0.0 for x in test]

        # Evaluate your predictions
        results = benchmark.evaluate(predictions)

        # Submit your results
        results.upload_to_hub(owner="dummy-user")
        ```

    Attributes:
        dataset: The dataset the benchmark specification is based on.
        target_cols: The column(s) of the original dataset that should be used as target.
        input_cols: The column(s) of the original dataset that should be used as input.
        split: The predefined train-test split to use for evaluation.
        metrics: The metrics to use for evaluating performance
        main_metric: The main metric used to rank methods. If `None`, the first of the `metrics` field.
        readme: Markdown text that can be used to provide a formatted description of the benchmark.
            If using the Polaris Hub, it is worth noting that this field is more easily edited through the Hub UI
            as it provides a rich text editor for writing markdown.
        target_types: A dictionary that maps target columns to their type. If not specified, this is automatically inferred.
    For additional meta-data attributes, see the [`BaseArtifactModel`][polaris._artifact.BaseArtifactModel] class.
    """

    # Public attributes
    # Data
    dataset: Union[DatasetV1, CompetitionDataset, str, dict[str, Any]]
    target_cols: ColumnsType
    input_cols: ColumnsType
    split: SplitType
    metrics: Union[str, Metric, list[str | Metric]]
    main_metric: str | Metric | None = None

    # Additional meta-data
    readme: str = ""
    target_types: dict[str, Union[TargetType, str, None]] = Field(default_factory=dict, validate_default=True)

    @field_validator("dataset")
    def _validate_dataset(cls, v):
        """
        Allows either passing a Dataset object or the kwargs to create one
        """
        if isinstance(v, dict):
            v = DatasetV1(**v)
        elif isinstance(v, str):
            v = DatasetV1.from_json(v)
        return v

    @field_validator("target_cols", "input_cols")
    def _validate_cols(cls, v, info: ValidationInfo):
        """Verifies all columns are present in the dataset."""
        if not isinstance(v, list):
            v = [v]
        if len(v) == 0:
            raise InvalidBenchmarkError("Specify at least a single column")
        if info.data.get("dataset") is not None and not all(
            c in info.data["dataset"].table.columns for c in v
        ):
            raise InvalidBenchmarkError("Not all specified target columns were found in the dataset.")
        return v

    @field_validator("metrics")
    def _validate_metrics(cls, v):
        """
        Verifies all specified metrics are either a Metric object or a valid metric name.
        Also verifies there are no duplicate metrics.

        If there are multiple test sets, it is assumed the same metrics are used across test sets.
        """
        if not isinstance(v, list):
            v = [v]

        v = [m if isinstance(m, Metric) else Metric[m] for m in v]

        if len(set(v)) != len(v):
            raise InvalidBenchmarkError("The task specifies duplicate metrics")

        if len(v) == 0:
            raise InvalidBenchmarkError("Specify at least one metric")

        return v

    @field_validator("main_metric")
    def _validate_main_metric(cls, v):
        """Converts the main metric to a Metric object if it is a string."""
        # lu: v can be None.
        if v and not isinstance(v, Metric):
            v = Metric[v]
        return v

    @model_validator(mode="after")
    def _validate_split(self) -> Self:
        """
        Verifies that:
          1) There are no empty test partitions
          2) All indices are valid given the dataset
          3) There is no duplicate indices in any of the sets
          4) There is no overlap between the train and test set
          5) No row exists in the test set where all labels are missing/empty
        """
        split = self.split

        # Train partition can be empty (zero-shot)
        # Test partitions cannot be empty
        if (isinstance(split[1], dict) and any(len(v) == 0 for v in split[1].values())) or (
            not isinstance(split[1], dict) and len(split[1]) == 0
        ):
            raise InvalidBenchmarkError("The predefined split contains empty test partitions")

        train_idx_list = split[0]
        full_test_idx_list = (
            list(chain.from_iterable(split[1].values())) if isinstance(split[1], dict) else split[1]
        )

        if len(train_idx_list) == 0:
            logger.info(
                "This benchmark only specifies a test set. It will return an empty train set in `get_train_test_split()`"
            )

        train_idx_set = set(train_idx_list)
        full_test_idx_set = set(full_test_idx_list)

        # The train and test indices do not overlap
        if len(train_idx_set & full_test_idx_set) > 0:
            raise InvalidBenchmarkError("The predefined split specifies overlapping train and test sets")

        # Check for duplicate indices within the train set
        if len(train_idx_set) != len(train_idx_list):
            raise InvalidBenchmarkError("The training set contains duplicate indices")

        # Check for duplicate indices within a given test set. Because a user can specify
        # multiple test sets for a given benchmark and it is acceptable for indices to be shared
        # across test sets, we check for duplicates in each test set independently.
        if isinstance(split[1], dict):
            for test_set_name, test_set_idx_list in split[1].items():
                if len(test_set_idx_list) != len(set(test_set_idx_list)):
                    raise InvalidBenchmarkError(
                        f'Test set with name "{test_set_name}" contains duplicate indices'
                    )
        elif len(full_test_idx_set) != len(full_test_idx_list):
            raise InvalidBenchmarkError("The test set contains duplicate indices")

        # All indices are valid given the dataset
        dataset = self.dataset
        if dataset is not None:
            max_i = len(dataset)
            if any(i < 0 or i >= max_i for i in chain(train_idx_list, full_test_idx_set)):
                raise InvalidBenchmarkError("The predefined split contains invalid indices")

        return self

    @field_validator("target_types")
    def _validate_target_types(cls, v, info: ValidationInfo):
        """Try to automatically infer the target types if not already set"""

        dataset = info.data.get("dataset")
        target_cols = info.data.get("target_cols")

        if dataset is None or target_cols is None:
            return v

        for target in target_cols:
            if target not in v:
                # Skip inferring the target type for pointer columns.
                # This would be complex to implement properly.
                # For these columns, dataset creators can still manually specify the target type.
                anno = dataset.annotations.get(target)
                if anno is not None and anno.is_pointer:
                    v[target] = None
                    continue

                val = dataset.table.loc[:, target]

                # Non numeric columns can be targets (e.g. prediction molecular reactions),
                # but in that case we currently don't infer the target type.
                if not np.issubdtype(val.dtype, np.number):
                    v[target] = None
                    continue

                # remove the nans for mutiple task dataset when the table is sparse
                target_type = type_of_target(val[~np.isnan(val)])
                if target_type == "continuous":
                    v[target] = TargetType.REGRESSION
                elif target_type in ["binary", "multiclass"]:
                    v[target] = TargetType.CLASSIFICATION
                else:
                    v[target] = None
            elif not isinstance(v, TargetType):
                v[target] = TargetType(v[target])
        return v

    @model_validator(mode="after")
    def _validate_model(self) -> Self:
        """
        Sets a default metric if missing.
        """
        # Set a default main metric if not set yet
        if self.main_metric is None:
            self.main_metric = self.metrics[0]
        return self

    @field_serializer("metrics", "main_metric")
    def _serialize_metrics(self, v):
        """Return the string identifier so we can serialize the object"""
        if isinstance(v, Metric):
            return v.name
        return [m.name for m in v]

    @field_serializer("split")
    def _serialize_split(self, v):
        """Convert any tuple to list to make sure it's serializable"""
        return listit(v)

    @field_serializer("target_types")
    def _serialize_target_types(self, v):
        """Convert from enum to string to make sure it's serializable"""
        return {k: v.value for k, v in self.target_types.items()}

    def _compute_checksum(self):
        """
        Computes a hash of the benchmark.

        This is meant to uniquely identify the benchmark and can be used to verify the version.
        """

        hash_fn = md5()
        hash_fn.update(self.dataset.md5sum.encode("utf-8"))
        for c in sorted(self.target_cols):
            hash_fn.update(c.encode("utf-8"))
        for c in sorted(self.input_cols):
            hash_fn.update(c.encode("utf-8"))
        for m in sorted(self.metrics, key=lambda k: k.name):
            hash_fn.update(m.name.encode("utf-8"))

        if not isinstance(self.split[1], dict):
            split = self.split[0], {"test": self.split[1]}
        else:
            split = self.split

        # Train set
        s = json.dumps(sorted(split[0]))
        hash_fn.update(s.encode("utf-8"))

        # Test sets
        for k in sorted(split[1].keys()):
            s = json.dumps(sorted(split[1][k]))
            hash_fn.update(k.encode("utf-8"))
            hash_fn.update(s.encode("utf-8"))

        checksum = hash_fn.hexdigest()
        return checksum

    @computed_field
    @property
    def n_train_datapoints(self) -> int:
        """The size of the train set."""
        return len(self.split[0])

    @computed_field
    @property
    def n_test_sets(self) -> int:
        """The number of test sets"""
        return len(self.split[1]) if isinstance(self.split[1], dict) else 1

    @computed_field
    @property
    def n_test_datapoints(self) -> dict[str, int]:
        """The size of (each of) the test set(s)."""
        if self.n_test_sets == 1:
            return {"test": len(self.split[1])}
        else:
            return {k: len(v) for k, v in self.split[1].items()}

    @computed_field
    @property
    def n_classes(self) -> dict[str, int]:
        """The number of classes for each of the target columns."""
        n_classes = {}
        for target in self.target_cols:
            target_type = self.target_types.get(target)
            if target_type is None or target_type == TargetType.REGRESSION:
                continue
            # TODO: Don't use table attribute
            n_classes[target] = self.dataset.table.loc[:, target].nunique()
        return n_classes

    @computed_field
    @property
    def task_type(self) -> str:
        """The high-level task type of the benchmark."""
        v = TaskType.MULTI_TASK if len(self.target_cols) > 1 else TaskType.SINGLE_TASK
        return v.value

    def _get_subset(self, indices, hide_targets=True, featurization_fn=None):
        """Returns a [`Subset`][polaris.dataset.Subset] using the given indices. Used
        internally to construct the train and test sets."""
        return Subset(
            dataset=self.dataset,
            indices=indices,
            input_cols=self.input_cols,
            target_cols=self.target_cols,
            hide_targets=hide_targets,
            featurization_fn=featurization_fn,
        )

    def _get_test_set(
        self, hide_targets=True, featurization_fn: Optional[Callable] = None
    ) -> Union["Subset", dict[str, Subset]]:
        """Construct the test set(s), given the split in the benchmark specification. Used
        internally to construct the test set for client use and evaluation.
        """

        def make_test_subset(vals):
            return self._get_subset(vals, hide_targets=hide_targets, featurization_fn=featurization_fn)

        test_split = self.split[1]
        if isinstance(test_split, dict):
            test = {k: make_test_subset(v) for k, v in test_split.items()}
        else:
            test = make_test_subset(test_split)

        return test

    def get_train_test_split(
        self, featurization_fn: Optional[Callable] = None
    ) -> tuple[Subset, Union["Subset", dict[str, Subset]]]:
        """Construct the train and test sets, given the split in the benchmark specification.

        Returns [`Subset`][polaris.dataset.Subset] objects, which offer several ways of accessing the data
        and can thus easily serve as a basis to build framework-specific (e.g. PyTorch, Tensorflow)
        data-loaders on top of.

        Args:
            featurization_fn: A function to apply to the input data. If a multi-input benchmark, this function
                expects an input in the format specified by the `input_format` parameter.

        Returns:
            A tuple with the train `Subset` and test `Subset` objects.
                If there are multiple test sets, these are returned in a dictionary and each test set has
                an associated name. The targets of the test set can not be accessed.
        """

        train = self._get_subset(self.split[0], hide_targets=False, featurization_fn=featurization_fn)
        test = self._get_test_set(hide_targets=True, featurization_fn=featurization_fn)

        return train, test

    def evaluate(
        self, y_pred: Optional[PredictionsType] = None, y_prob: Optional[PredictionsType] = None
    ) -> BenchmarkResults:
        """Execute the evaluation protocol for the benchmark, given a set of predictions.

        info: What about `y_true`?
            Contrary to other frameworks that you might be familiar with, we opted for a signature that includes just
            the predictions. This reduces the chance of accidentally using the test targets during training.

        info: Expected structure for `y_pred` and `y_prob` arguments
            The supplied `y_pred` and `y_prob` arguments must adhere to a certain structure depending on the number of
            tasks and test sets included in the benchmark. Refer to the following for guidance on the correct structure when
            creating your `y_pred` and `y_prod` objects:

            - Single task, single set: `[values...]`
            - Multi-task, single set: `{task_name_1: [values...], task_name_2: [values...]}`
            - Single task, multi-set: `{test_set_1: {task_name: [values...]}, test_set_2: {task_name: [values...]}}`
            - Multi-task, multi-set: `{test_set_1: {task_name_1: [values...], task_name_2: [values...]}, test_set_2: {task_name_1: [values...], task_name_2: [values...]}}`

        For this method, we make the following assumptions:

        1. There can be one or multiple test set(s);
        2. There can be one or multiple target(s);
        3. The metrics are _constant_ across test sets;
        4. The metrics are _constant_ across targets;
        5. There can be metrics which measure across tasks.

        Args:
            y_pred: The predictions for the test set, as NumPy arrays.
                If there are multiple targets, the predictions should be wrapped in a dictionary with the target labels as keys.
                If there are multiple test sets, the predictions should be further wrapped in a dictionary
                    with the test subset labels as keys.
            y_prob: The predicted probabilities for the test set, as NumPy arrays.

        Returns:
            A `BenchmarkResults` object. This object can be directly submitted to the Polaris Hub.

        Examples:
            1. For regression benchmarks:
                pred_scores = your_model.predict_score(molecules) # predict continuous score values
                benchmark.evaluate(y_pred=pred_scores)
            2. For classification benchmarks:
                - If `roc_auc` and `pr_auc` are in the metric list, both class probabilities and label predictions are required:
                    pred_probs = your_model.predict_proba(molecules) # predict probablities
                    pred_labels = your_model.predict_labels(molecules) # predict class labels
                    benchmark.evaluate(y_pred=pred_labels, y_prob=pred_probs)
                - Otherwise:
                    benchmark.evaluate(y_pred=pred_labels)
        """

        # Instead of having the user pass the ground truth, we extract it from the benchmark spec ourselves.
        # The `evaluate_benchmark` function expects the benchmark labels to be of a certain structure which
        # depends on the number of tasks and test sets defined for the benchmark. Below, we build the structure
        # of the benchmark labels based on the aforementioned factors.
        test = self._get_test_set(hide_targets=False)
        if isinstance(test, dict):
            #
            # For multi-set benchmarks
            y_true = {}
            for test_set_name, values in test.items():
                y_true[test_set_name] = {}
                if isinstance(values.targets, dict):
                    #
                    # For multi-task, multi-set benchmarks
                    for task_name, values in values.targets.items():
                        y_true[test_set_name][task_name] = values
                else:
                    #
                    # For single task, multi-set benchmarks
                    y_true[test_set_name][self.target_cols[0]] = values.targets
        else:
            #
            # For single set benchmarks (single and multiple task)
            y_true = test.targets

        scores = evaluate_benchmark(self.target_cols, self.metrics, y_true, y_pred=y_pred, y_prob=y_prob)

        return BenchmarkResults(results=scores, benchmark_name=self.name, benchmark_owner=self.owner)

    def upload_to_hub(
        self,
        settings: Optional[PolarisHubSettings] = None,
        cache_auth_token: bool = True,
        access: Optional[AccessType] = "private",
        owner: Union[HubOwner, str, None] = None,
        **kwargs: dict,
    ):
        """
        Very light, convenient wrapper around the
        [`PolarisHubClient.upload_benchmark`][polaris.hub.client.PolarisHubClient.upload_benchmark] method.
        """
        from polaris.hub.client import PolarisHubClient

        with PolarisHubClient(
            settings=settings,
            cache_auth_token=cache_auth_token,
            **kwargs,
        ) as client:
            return client.upload_benchmark(self, access=access, owner=owner)

    def to_json(self, destination: str) -> str:
        """Save the benchmark to a destination directory as a JSON file.

        Warning: Multiple files
            Perhaps unintuitive, this method creates multiple files in the destination directory as it also saves
            the dataset it is based on to the specified destination.
            See the docstring of [`Dataset.to_json`][polaris.dataset.Dataset.to_json] for more information.

        Args:
            destination: The _directory_ to save the associated data to.

        Returns:
            The path to the JSON file.
        """

        fs.mkdir(destination, exist_ok=True)

        data = self.model_dump()
        data["dataset"] = self.dataset.to_json(destination=destination)

        path = fs.join(destination, "benchmark.json")
        with fsspec.open(path, "w") as f:
            json.dump(data, f)

        return path

    def _repr_dict_(self) -> dict:
        """Utility function for pretty-printing to the command line and jupyter notebooks"""
        repr_dict = self.model_dump()
        repr_dict.pop("dataset")
        repr_dict.pop("split")
        repr_dict["dataset_name"] = self.dataset.name
        return repr_dict

    def _repr_html_(self):
        """For pretty printing in Jupyter."""
        return dict2html(self._repr_dict_())

    def __repr__(self):
        return json.dumps(self._repr_dict_(), indent=2)

    def __str__(self):
        return self.__repr__()

    def __eq__(self, other):
        if not isinstance(other, BenchmarkSpecification):
            return False
        return self.md5sum == other.md5sum



from pydantic import model_validator
from typing_extensions import Self

from polaris.dataset._dataset import DatasetV1
from polaris.utils.errors import InvalidCompetitionError


class CompetitionDataset(DatasetV1):
    """Dataset subclass for Polaris competitions.

    In addition to the data model and logic of the base Dataset class,
    this class adds additional functionality which validates certain aspects
    of the training data for a given competition.
    """

    @model_validator(mode="after")
    def _validate_model(self) -> Self:
        """We reject the instantiation of competition datasets which leverage Zarr for the time being"""

        if self.uses_zarr:
            raise InvalidCompetitionError("Pointer columns are not currently supported in competitions.")

        return self



import abc
from typing import Dict, Tuple, TypeAlias, Union

import pandas as pd

from polaris.dataset import ColumnAnnotation
from polaris.dataset._adapters import Adapter
from polaris.dataset._dataset import _INDEX_SEP

FactoryProduct: TypeAlias = Tuple[pd.DataFrame, Dict[str, ColumnAnnotation], Dict[str, Adapter]]


class Converter(abc.ABC):
    @abc.abstractmethod
    def convert(self, path: str, append: bool = False) -> FactoryProduct:
        """This converts a file into a table and possibly annotations"""
        raise NotImplementedError

    @staticmethod
    def get_pointer(column: str, index: Union[int, slice]) -> str:
        """
        Creates a pointer.

        Args:
            column: The name of the column. Each column has its own group in the root.
            index: The index or slice of the pointer.
        """
        if isinstance(index, slice):
            index_substr = f"{_INDEX_SEP}{index.start}:{index.stop}"
        else:
            index_substr = f"{_INDEX_SEP}{index}"
        return f"{column}{index_substr}"



from typing import Union
from urllib.parse import urljoin

from pydantic import ValidationInfo, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict

from polaris.utils.types import HttpUrlString, TimeoutTypes


class PolarisHubSettings(BaseSettings):
    """Settings for the OAuth2 Polaris Hub API Client.

    Info: Secrecy of these settings
        Since the Polaris Hub uses PCKE (Proof Key for Code Exchange) for OAuth2,
        these values thus do not have to be kept secret.
        See [RFC 7636](https://datatracker.ietf.org/doc/html/rfc7636) for more info.

    Attributes:
        hub_url: The URL to the main page of the Polaris Hub.
        api_url: The URL to the main entrypoint of the Polaris API.
        authorize_url: The URL of the OAuth2 authorization endpoint.
        callback_url: The URL to which the user is redirected after authorization.
        token_fetch_url: The URL of the OAuth2 token endpoint.
        user_info_url: The URL of the OAuth2 user info endpoint.
        scopes: The OAuth2 scopes that are requested.
        client_id: The OAuth2 client ID.
        ca_bundle: The path to a CA bundle file for requests.
            Allows for custom SSL certificates to be used.
    """

    # Configuration of the pydantic model
    model_config = SettingsConfigDict(env_file=".env", env_prefix="POLARIS_")

    # Hub settings
    hub_url: HttpUrlString = "https://polarishub.io/"
    api_url: HttpUrlString | None = None
    custom_metadata_prefix: str = "X-Amz-Meta-"

    # Hub authentication settings
    hub_token_url: HttpUrlString | None = None

    # External authentication settings
    authorize_url: HttpUrlString = "https://clerk.polarishub.io/oauth/authorize"
    callback_url: HttpUrlString = "https://polarishub.io/oauth2/callback"
    token_fetch_url: HttpUrlString = "https://clerk.polarishub.io/oauth/token"
    user_info_url: HttpUrlString = "https://clerk.polarishub.io/oauth/userinfo"
    scopes: str = "profile email"
    client_id: str = "agQP2xVM6JqMHvGc"

    # Networking settings
    ca_bundle: Union[str, bool, None] = None
    default_timeout: TimeoutTypes = (10, 200)

    @field_validator("api_url", mode="before")
    def validate_api_url(cls, v, info: ValidationInfo):
        if v is None:
            v = urljoin(str(info.data["hub_url"]), "/api")
        return v

    @field_validator("hub_token_url", mode="before")
    def populate_hub_token_url(cls, v, info: ValidationInfo):
        if v is None:
            v = urljoin(str(info.data["hub_url"]), "/api/auth/token")
        return v
